unit DXSound1;

interface

uses
   Windows, SysUtils, Classes, Dialogs, mmSystem,
   DSound, D3DTypes;

const
   MAX_CHANNEL       = 20;

   DS_BITPERSAMPLE   = 8;
   DS_CHANSPERSAMPLE = 1;
   DS_SAMPLERATE     = 22050;

   NUM_BUFFER_SEGMENTS = 6;
   PLAYBACK_TIMER_PERIOD = 200;


type
   EBufferAttribute = (ba2DBuffer, ba3DBuffer);

   TDXSound = class

      constructor Create(Handle : HWnd);
      procedure   Free;

   private
      ParentHandle         : HWnd;

      procedure SetVolume(Channel : integer; Value : Longint);
      function  GetVolume(Channel : integer) : Longint;
      procedure SetPan(Channel : integer; Value : Longint);
      function  GetPan(Channel : integer) : Longint;
      procedure Set3D(Channel : integer; D3DPos : TD3DVECTOR);

   public
      DirectSound          : IDirectSound;

      Buffer               : array[0..MAX_CHANNEL] of PByte;
      BufferLength         : array[0..MAX_CHANNEL] of DWORD;
      DirectSoundBuffer    : array[0..MAX_CHANNEL] of IDirectSoundBuffer;
      DirectStreamBuffer   : IDirectSoundBuffer;
      DirectSound3DBuffer  : array[0..MAX_CHANNEL] of IDirectSound3DBuffer;
      SoundBufferAttribute : array[0..MAX_CHANNEL] of EBufferAttribute;

      ActiveChannel        : integer;

      Stream_mmIO          : HmmIO;
      Stream_WaveFormatEx  : TWaveFormatEx;
      Stream_BufferSize    : DWORD;

      function    CreatePrimaryBuffer : boolean;
      function    CreateSoundBuffer(Channel : integer; WaveFormatEx  : TWaveFormatEx; Is3DBuffer : Boolean) : boolean;
      function    AllocateChannel : integer;
      function    FillBuffer(Channel : integer) : boolean;
//      procedure   FillStreamBuffer;

      function    LoadWaveResource(Instance : HINST; ResName : String; RestType : PChar; Is3DBuffer : Boolean) : integer;
      function    LoadWave(FileName : PChar; Is3DBuffer : Boolean) : integer;
      procedure   Play(ActiveChannel : integer);
      procedure   Play3D(ActiveChannel : integer; D3DPos : TD3DVECTOR);
      procedure   Stop(ActiveChannel : integer);

//      function    CreateStreamWave(FileName : PChar) : Boolean;
//      procedure   PlayStreamWave;
//      procedure   StopStreamWave;
//      procedure   DestroyStreamBuffer;

      // Range : 0 - 256
      property    Volume[Channel : integer] : Longint read  GetVolume write SetVolume;
      // 256 -> object locates LEFT, -256 -> object locates RIGHT
      property    Pan   [Channel : integer] : Longint read  GetPan    write SetPan;
      property    Pos3D [Channel : integer] : TD3DVECTOR              write Set3D;
   end;

implementation

const
   MAX_PAN_VALUE   = 256;
   MIN_PAN_VALUE   = -256;
   MIN_VOL_VALUE   = 0;
   MAX_VOL_VALUE   = 256;
   SHIFT_VOL_VALUE = -256;
   SHIFT_PAN_VALUE = -0;
   MUL_PAN_VALUE   = 20;
   MUL_VOL_VALUE   = 20;

constructor TDXSound.Create(Handle : HWnd);
var
   i        : integer;
   DSCaps   : TDSCaps;
   DSResult : HResult;
begin
   inherited Create;

   ParentHandle  := Handle;
   ActiveChannel := -1;

   DirectSound := nil;
   for i := 0 to MAX_CHANNEL do begin
      Buffer              [i] := nil;
      DirectSoundBuffer   [i] := nil;
      DirectSound3DBuffer [i] := nil;
      SoundBufferAttribute[i] := ba2DBuffer;
   end;

   DirectSoundCreate(nil,DirectSound,nil);
   if not Assigned(DirectSound) then exit;

   FillChar(DSCaps,SizeOf(DSCaps),0);
   DSCaps.dwSize := sizeof(TDSCaps);
   DSResult := DirectSound.GetCaps(DSCaps);
   if DSResult = DS_OK then begin
   end;

   DirectSound.SetCooperativeLevel(ParentHandle,DSSCL_NORMAL);

end;

procedure   TDXSound.Free;
var
   i : integer;
begin
   for i := 0 to MAX_CHANNEL do begin
      if Assigned(Buffer[i]) then begin
         FreeMem(Buffer[i],BufferLength[i]);
         Buffer[i] := nil;
      end;
      if Assigned(DirectSoundBuffer[i]) then begin
         DirectSoundBuffer[i]._Release;
         Pointer(DirectSoundBuffer[i]) := nil;
      end;
      if Assigned(DirectSound3DBuffer[i]) then begin
         DirectSound3DBuffer[i]._Release;
         Pointer(DirectSound3DBuffer[i]) := nil;
      end;
   end;

   if Assigned(DirectSound) then begin
      DirectSound._Release;
      Pointer(DirectSound) := nil;
   end;

   inherited Free;
end;

procedure   TDXSound.SetVolume(Channel : integer; Value : Longint);
var
   HR : Longint;
begin
   if not Assigned(DirectSound) then exit;

   if Channel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[Channel]) then begin
      if Value < MIN_VOL_VALUE then Value := MIN_VOL_VALUE;
      if Value > MAX_VOL_VALUE then Value := MAX_VOL_VALUE;
      HR := DirectSoundBuffer[Channel].SetVolume((Value + SHIFT_VOL_VALUE) * MUL_VOL_VALUE);
      if HR <> 0 then ShowMessage(IntToStr(HR));
   end;
end;

function    TDXSound.GetVolume(Channel : integer) : Longint;
var
   HR    : Longint;
   Value : Longint;
begin
   GetVolume := 0;
   if not Assigned(DirectSound) then exit;

   if Channel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[Channel]) then begin
      Value := 0;
      HR := DirectSoundBuffer[Channel].GetVolume(Value);
      if HR <> 0 then ShowMessage(IntToStr(HR-Longint($88780000)));
      GetVolume := Value div MUL_VOL_VALUE - SHIFT_VOL_VALUE;
   end;
end;

procedure   TDXSound.SetPan(Channel : integer; Value : Longint);
var
   HR : Longint;
begin
   if not Assigned(DirectSound) then exit;

   if Channel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[Channel]) then begin
      if Value < MIN_PAN_VALUE then Value := MIN_PAN_VALUE;
      if Value > MAX_PAN_VALUE then Value := MAX_PAN_VALUE;
      HR := DirectSoundBuffer[Channel].SetPan((Value + SHIFT_PAN_VALUE) * MUL_PAN_VALUE);
      if HR <> 0 then ShowMessage(IntToStr(HR));
   end;
end;

function    TDXSound.GetPan(Channel : integer) : Longint;
var
   HR    : Longint;
   Value : Longint;
begin
   GetPan := 0;
   if not Assigned(DirectSound) then exit;

   if Channel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[Channel]) then begin
      Value := 0;
      HR := DirectSoundBuffer[Channel].GetPan(Value);
      if HR <> 0 then ShowMessage(IntToStr(HR-Longint($88780000)));
      GetPan := Value div MUL_PAN_VALUE - SHIFT_PAN_VALUE;
   end;
end;

procedure   TDXSound.Set3D(Channel : integer; D3DPos : TD3DVECTOR);
begin
   DirectSound3DBuffer[Channel].SetPosition(D3DPos.x,D3DPos.y,D3DPos.z,DS3D_IMMEDIATE);
end;

function    TDXSound.CreatePrimaryBuffer : boolean;
var
   DSBufferDesc  : TDSBUFFERDESC;
   WaveFormatEx  : TWaveFormatEx;
   DSBCaps       : TDSBCaps;
   DSResult      : HResult;
begin
   CreatePrimaryBuffer := FALSE;
   if not Assigned(DirectSound) then exit;

   FillChar(WaveFormatEx,sizeof(TWaveFormatEx),0);

   WaveFormatEx.wFormatTag      := WAVE_FORMAT_PCM;
   WaveFormatEx.nChannels       := DS_CHANSPERSAMPLE;
   WaveFormatEx.nSamplesPerSec  := DS_SAMPLERATE;
   WaveFormatEx.nBlockAlign     := WaveFormatEx.nChannels * DS_BITPERSAMPLE div 8;
   WaveFormatEx.nAvgBytesPerSec := WaveFormatEx.nSamplesPerSec * WaveFormatEx.nBlockAlign;
   WaveFormatEx.wBitsPerSample  := DS_BITPERSAMPLE;
   WaveFormatEx.cbSize          := 0;

   FillChar(DSBufferDesc,sizeof(TDSBufferDesc),0);

   DSBufferDesc.dwSize          := sizeof(TDSBufferDesc);
   DSBufferDesc.dwFlags         := DSBCAPS_PRIMARYBUFFER;

   DSResult := DirectSound.SetCooperativeLevel(ParentHandle,DSSCL_WRITEPRIMARY);
   if DSResult = DS_OK then begin
      DSResult := DirectSound.CreateSoundBuffer(DSBufferDesc,DirectSoundBuffer[0],nil);
      if DSResult = DS_OK then begin
         DSResult := DirectSoundBuffer[0].SetFormat(WaveFormatEx);
         if DSResult = DS_OK then begin
            DSBCaps.dwSize      := sizeof(TDSBCaps);
            DirectSoundBuffer[0].GetCaps(DSBCaps);
            BufferLength[0]     := DSBCaps.dwBufferBytes;
            CreatePrimaryBuffer := TRUE;
            exit;
         end;
      end;
   end;

   DirectSoundBuffer[0] := nil;
   BufferLength[0]      := 0;
end;

function    TDXSound.CreateSoundBuffer(Channel : integer; WaveFormatEx  : TWaveFormatEx; Is3DBuffer : Boolean) : boolean;
var
   DSBufferDesc  : TDSBUFFERDESC;
   DSResult      : HResult;
begin
   CreateSoundBuffer := FALSE;
   if not Assigned(DirectSound) then exit;

   if Channel = 0 then begin
      CreatePrimaryBuffer;
      CreateSoundBuffer := TRUE;
      Exit;
   end;

   if not (Channel in [1..MAX_CHANNEL]) then exit;
{
   FillChar(WaveFormatEx,sizeof(TWaveFormatEx),0);

   WaveFormatEx.wFormatTag      := WAVE_FORMAT_PCM;
   WaveFormatEx.nChannels       := DS_CHANSPERSAMPLE;
   WaveFormatEx.nSamplesPerSec  := DS_SAMPLERATE;
   WaveFormatEx.nBlockAlign     := WaveFormatEx.nChannels * DS_BITPERSAMPLE div 8;
   WaveFormatEx.nAvgBytesPerSec := WaveFormatEx.nSamplesPerSec * WaveFormatEx.nBlockAlign;
   WaveFormatEx.wBitsPerSample  := DS_BITPERSAMPLE;
}
   WaveFormatEx.cbSize          := sizeof(TWaveFormatEx);

   FillChar(DSBufferDesc,sizeof(TDSBufferDesc),0);

   DSBufferDesc.dwSize          := sizeof(TDSBufferDesc);
   if not Is3DBuffer then
      DSBufferDesc.dwFlags      := DSBCAPS_CTRLDEFAULT or DSBCAPS_STATIC or DSBCAPS_GETCURRENTPOSITION2
   else
      DSBufferDesc.dwFlags      := DSBCAPS_CTRL3D;
   DSBufferDesc.dwBufferBytes   := BufferLength[Channel];
   DSBufferDesc.lpwfxFormat     := @WaveFormatEx;

   DSResult := DirectSound.CreateSoundBuffer(DSBufferDesc,DirectSoundBuffer[Channel],nil);

   if not Is3DBuffer then begin
      SoundBufferAttribute[Channel] := ba2DBuffer
   end else begin
      SoundBufferAttribute[Channel] := ba3DBuffer;
      DirectSoundBuffer[Channel].QueryInterface(IID_IDirectSound3DBuffer,DirectSound3DBuffer[Channel]);
   end;

   if DSResult <> DS_OK then begin
      Raise Exception.Create(DSound.ErrorString(DSResult));
      if DSResult = DSERR_BADFORMAT then
         Raise Exception.Create(' ʴ Wave Դϴ.');
      DirectSoundBuffer[Channel] := nil;
      exit;
   end;

   CreateSoundBuffer := TRUE;
end;

function    TDXSound.AllocateChannel : integer;
var
   i : word;
begin
   AllocateChannel := -1;
   if not Assigned(DirectSound) then exit;

   for i := 1 to MAX_CHANNEL do begin
      if not Assigned(DirectSoundBuffer[i]) then begin
         AllocateChannel := i;
         exit;
      end;
   end;
end;

function    TDXSound.FillBuffer(Channel : integer) : boolean;
var
   AudioPtr1,   AudioPtr2   : Pointer;
   AudioBytes1, AudioBytes2 : DWORD;
   DSResult                 : HResult;
begin
   FillBuffer := FALSE;
   if not Assigned(DirectSound) then exit;

   DSResult   := DS_OK;
   repeat
      if DSResult = DSERR_BUFFERLOST then DirectSoundBuffer[Channel].Restore;
      DSResult := DirectSoundBuffer[Channel].Lock(0,BufferLength[Channel],AudioPtr1,AudioBytes1,AudioPtr2,AudioBytes2,0);
   until DSResult <> DSERR_BUFFERLOST;

   if DSResult <> DS_OK then exit;

   CopyMemory(AudioPtr1,Buffer[Channel],AudioBytes1);
   if Assigned(AudioPtr2) then begin
      CopyMemory(AudioPtr2,PByte(DWORD(Buffer[Channel])+AudioBytes1),AudioBytes2);
   end;

   DSResult := DirectSoundBuffer[Channel].UnLock(AudioPtr1,AudioBytes1,AudioPtr2,AudioBytes2);

   if DSResult <> DS_OK then exit;

   FillBuffer := TRUE;
end;

function    mmioFOURCC(Data1,Data2,Data3,Data4 : char) : FOURCC;
var
   Buffer : FOURCC;
begin
   Buffer := ord(Data4);
   Buffer := Buffer shl 8;
   Buffer := Buffer or ord(Data3);
   Buffer := Buffer shl 8;
   Buffer := Buffer or ord(Data2);
   Buffer := Buffer shl 8;
   Buffer := Buffer or ord(Data1);

   mmioFOURCC := Buffer;
end;

function    TDXSound.LoadWaveResource(Instance : HINST; ResName : String; RestType : PChar; Is3DBuffer : Boolean) : integer;
var
   WaveFormatEx  : TWaveFormatEx;
   WaveSize      : DWORD;
   Channel       : integer;
   Resource      : TResourceStream;
begin
   LoadWaveResource := -1;
   if not Assigned(DirectSound) then exit;

   Resource := TResourceStream.Create(Instance,ResName,RestType);

   // RIFF(4) ü(4) WAVE(4) fmt_(4)
   Resource.Seek(16,soFromBeginning);
   // WaveFormatEx(4)
   Resource.Read(WaveSize,Sizeof(WaveSize));
   // WaveFormatEx(16-18)
   Resource.Read(WaveFormatEx,WaveSize);
   // data(4)
   Resource.Seek( 4,soFromCurrent);
   // ͱ(4)
   Resource.Read(WaveSize,Sizeof(WaveSize));

   Channel := AllocateChannel;

   if Channel < 0 then begin
      exit;               //   ä .
   end;

   GetMem(Buffer[Channel],WaveSize);
   BufferLength[Channel] := WaveSize;

   if not Assigned(Buffer[Channel]) then exit;

   if Resource.Read(Buffer[Channel]^,WaveSize) <> WaveSize then begin
      Resource.Free;
      exit;               // data ûũ б .
   end;

   Resource.Free;

   if not CreateSoundBuffer(Channel,WaveFormatEx,Is3DBuffer) then begin
      exit;               // Sound Buffer  
   end;

   FillBuffer(Channel);

   ActiveChannel    := Channel;
   LoadWaveResource := Channel;
end;

function    TDXSound.LoadWave(FileName : PChar; Is3DBuffer : Boolean) : integer;
var
   mmIO          : HmmIO;
   Parent, Child : TmmCkInfo;
   WaveFormatEx  : TWaveFormatEx;
   Channel       : integer;
begin
   LoadWave := -1;
   if not Assigned(DirectSound) then exit;

   FillChar(Parent,sizeof(TmmCkInfo),0);
   FillChar(Child ,sizeof(TmmCkInfo),0);

   mmIO  := mmioOpen(FileName,nil,MMIO_READ or MMIO_ALLOCBUF);

   if mmIO = 0 then exit;

   Parent.fccType := mmioFOURCC('W','A','V','E');
   if mmioDescend(mmIO,@Parent,nil,MMIO_FINDRIFF) <> 0 then begin
      mmioClose(mmIO,0);
      exit;               // Wav  ƴϴ.
   end;

   Child.ckid := mmioFOURCC('f','m','t',' ');
   if mmioDescend(mmIO,@Child,@Parent,0) <> 0 then begin
      mmioClose(mmIO,0);
      exit;               // fmt ûũ .
   end;

   if mmIORead(mmIO,PChar(@WaveFormatEx),sizeof(TWaveFormatEx)) <> sizeof(TWaveFormatEx) then begin
      mmioClose(mmIO,0);
      exit;               // fmt ûũ б .
   end;

   if WaveFormatEx.wFormatTag <> WAVE_FORMAT_PCM then begin
      mmioClose(mmIO,0);
      exit;               // PCM  ƴϴ.
   end;

   if mmioAscend(mmIO,@Child,0) <> 0 then begin
      mmioClose(mmIO,0);
      exit;               // ܰ ö  .
   end;

   Child.ckid := mmioFOURCC('d','a','t','a');
   if mmioDescend(mmIO,@Child,@Parent,MMIO_FINDCHUNK) <> 0 then begin
      mmioClose(mmIO,0);
      exit;               // data ûũ .
   end;

   Channel := AllocateChannel;

   if Channel < 0 then begin
      exit;               //   ä .
   end;

   GetMem(Buffer[Channel],Child.cksize);
   BufferLength[Channel] := Child.cksize;

   if not Assigned(Buffer[Channel]) then exit;

   if mmIORead(mmIO,PChar(Buffer[Channel]),Child.cksize) <> Child.cksize then begin
      mmioClose(mmIO,0);
      exit;               // data ûũ б .
   end;

   mmIOClose(mmIO,0);

   if not CreateSoundBuffer(Channel,WaveFormatEx,Is3DBuffer) then begin
      exit;               // Sound Buffer  
   end;

   FillBuffer(Channel);

   ActiveChannel := Channel;
   LoadWave      := Channel;
end;

procedure   TDXSound.Play(ActiveChannel : integer);
begin
   if not Assigned(DirectSound) then exit;

   if ActiveChannel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[ActiveChannel]) then begin
      Stop(ActiveChannel);
      DirectSoundBuffer[ActiveChannel].SetCurrentPosition(0);
      DirectSoundBuffer[ActiveChannel].Play(0,0,0);
   end;
end;

procedure   TDXSound.Play3D(ActiveChannel : integer; D3DPos : TD3DVECTOR);
begin
   if not Assigned(DirectSound) then exit;

   if ActiveChannel in [1..MAX_CHANNEL] then
   if Assigned(DirectSound3DBuffer[ActiveChannel]) then begin
      Stop(ActiveChannel);
      DirectSound3DBuffer[ActiveChannel].SetPosition(D3DPos.x,D3DPos.y,D3DPos.z,DS3D_IMMEDIATE);
      DirectSoundBuffer[ActiveChannel].SetCurrentPosition(0);
      DirectSoundBuffer[ActiveChannel].Play(0,0,0);
//      Play(ActiveChannel);
   end;
end;

procedure   TDXSound.Stop(ActiveChannel : integer);
begin
   if not Assigned(DirectSound) then exit;

   if ActiveChannel in [1..MAX_CHANNEL] then
   if Assigned(DirectSoundBuffer[ActiveChannel]) then
	  DirectSoundBuffer[ActiveChannel].Stop;
end;

end.
