unit SbModule;

interface

uses
   Windows, Classes,
   DDraw, MainUtDX, Module1;

type
   TSpriteMode   = (smCentered, smLeftTop);
   TColorEffect  = (ceNone, ceAdd, ceRed, ceGreen, ceBlue);

var
   MakeRGB    : function  (_R, _G, _B : Byte) : Word;
   Color2RGB  : procedure (Color : Word; var R, G, B : Byte);
   RGB2Color  : procedure (R, G, B : Byte; var Color : Word);
   PutPixel   : procedure (x, y : Longint; Color : Word);

   MIN_X_BOUNDARY : integer;
   MAX_X_BOUNDARY : integer;
   MIN_Y_BOUNDARY : integer;
   MAX_Y_BOUNDARY : integer;

   FlipPerSecond  : Double;     // ʴ  ȸ

 procedure SetWriteMode(PixelFormat : TPixelFormat);
 function  FixColor(Color : Word) : Word;

 procedure PutEffect(DestX, DestY : integer;
                     SourRect     : TRect;
                     Sour, Dest   : IDirectDrawSurface7;
                     Ratio        : Integer;
                     ColorEffect  : TColorEffect);

 procedure putBrightImage(DestX, DestY : integer;
                          SourRect     : TRect;
                          Sour, Dest   : IDirectDrawSurface7;
                          Ratio        : Integer);

 procedure TransparentCopy50(DestX, DestY : Integer;
                             SourX, SourY : Integer;
                             Size         : TPoint;
                             Sour, Dest   : IDirectDrawSurface7);

 procedure FadeCopy(Sour, Dest : IDirectDrawSurface7; FadeRatio : integer);

 procedure ModifyEdge(var x, y : integer; var ARect : TRect; Mode : TSpriteMode);
 procedure PutSprite(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);
 procedure PutSpriteMesh(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode; Ratio : integer);
 procedure PutSpriteTrans(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);
 procedure PutSpriteTrans75(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);

 procedure FlipPage;

implementation

procedure PutEffectVBClip(DestX, DestY : integer;
                          SourRect     : TRect;
                          Sour, Dest   : IDirectDrawSurface7;
                          Ratio        : Integer);
var
   R, G, B, _R, _G, _B        : Byte;
   Data                       : Word;
   x, y                       : Integer;
   SourX, SourY, SizeX, SizeY : Integer;
   SourDesc, DestDesc         : TDDSurfaceDesc2;
   pSour, pDest               : PWord;
   Red, Green, Blue           : Smallint;
begin

   SourX := SourRect.Left;
   SourY := SourRect.Top;
   SizeX := SourRect.Right  - SourRect.Left;
   SizeY := SourRect.Bottom - SourRect.Top;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Sour.Lock(nil,SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Dest.Lock(nil,DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   for y := 0 to Pred(SizeY) do begin
      for x := 0 to Pred(SizeX) do begin

         if (DestX+x < MIN_X_BOUNDARY) or (DestX+x >= MAX_X_BOUNDARY) or
            (DestY+y < MIN_Y_BOUNDARY) or (DestY+y >= MAX_Y_BOUNDARY) then Continue;

         pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX+x)*BPP);
         pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX+x)*BPP);
         Data := pSour^;

         if Data <> 0 then begin
            Color2RGB(pDest^,R,G,B);
            Color2RGB(Data,_R,_G,_B);

            Red   := Word(_R) * Ratio div 128 + R;
            Green := Word(_G) * Ratio div 128 + G;
            Blue  := Word(_B) * Ratio div 128 + B;

            if Red   < 0   then Red   := 0;
            if Red   > $1F then Red   := $1F;
            if Green < 0   then Green := 0;
            if Green > $1F then Green := $1F;
            if Blue  < 0   then Blue  := 0;
            if Blue  > $1F then Blue  := $1F;

            RGB2Color(Red,Green,Blue,Data);
            pDest^ := Data;
         end;
      end;
   end;

   Sour.UnLock(SourDesc.lpSurface);
   Dest.UnLock(DestDesc.lpSurface);
end;

procedure PutEffect(DestX, DestY : integer;
                    SourRect     : TRect;
                    Sour, Dest   : IDirectDrawSurface7;
                    Ratio        : Integer;
                    ColorEffect  : TColorEffect);
var
   R, G, B, _R, _G, _B        : Byte;
   Data                       : Word;
   x, y                       : Integer;
   SourX, SourY, SizeX, SizeY : Integer;
   SizeXHalf, SizeYHalf       : Integer;
   SourDesc, DestDesc         : TDDSurfaceDesc2;
   pSour, pDest               : PWord;
   Red, Green, Blue           : Smallint;
   tr1, tr2                   : Longint;
   tr3, Division              : Word;
begin
   if (not Assigned(Sour)) or (not Assigned(Dest)) then Exit;

   SourX     := SourRect.Left;
   SourY     := SourRect.Top;
   SizeX     := SourRect.Right  - SourRect.Left;
   SizeY     := SourRect.Bottom - SourRect.Top;
   SizeXHalf := SizeX div 2;
   SizeYHalf := SizeY div 2;

   if (DestX < MIN_X_BOUNDARY) or (DestX+SizeX >= MAX_X_BOUNDARY) or
      (DestY < MIN_Y_BOUNDARY) or (DestY+SizeX >= MAX_Y_BOUNDARY) then begin
      PutEffectVBClip(DestX,DestY,SourRect,Sour,Dest,Ratio);
      Exit;
   end;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Sour.Lock(nil,SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Dest.Lock(nil,DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   Division := SizeX * SizeY div Ratio div 3;

   case ColorEffect of
      ceNone  :
      begin
         if Ratio <   0 then Ratio :=   0;
         if Ratio > 128 then Ratio := 128;
         for y := 0 to Pred(SizeY) do begin
            pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
            pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
            for x := 0 to Pred(SizeX) do begin
               Data := pSour^;
               if Data <> 0 then begin
                  Color2RGB(pDest^,R,G,B);
                  Color2RGB(Data,_R,_G,_B);

                  Red   := (Word(_R) * Ratio + Word(R) * (128-Ratio)) div 128;
                  Green := (Word(_G) * Ratio + Word(G) * (128-Ratio)) div 128;
                  Blue  := (Word(_B) * Ratio + Word(B) * (128-Ratio)) div 128;

                  if Red   < 0   then Red   := 0;
                  if Red   > $1F then Red   := $1F;
                  if Green < 0   then Green := 0;
                  if Green > $1F then Green := $1F;
                  if Blue  < 0   then Blue  := 0;
                  if Blue  > $1F then Blue  := $1F;

                  RGB2Color(Red,Green,Blue,Data);
                  pDest^ := Data;
               end;
               inc(pSour);
               inc(pDest);
            end;
         end;
      end;
      ceAdd   :
      begin
         for y := 0 to Pred(SizeY) do begin
            pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
            pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
            for x := 0 to Pred(SizeX) do begin
               Data := pSour^;
               if Data <> 0 then begin
                  Color2RGB(pDest^,R,G,B);
                  Color2RGB(Data,_R,_G,_B);

                  Red   := Word(_R) * Ratio div 128 + R;
                  Green := Word(_G) * Ratio div 128 + G;
                  Blue  := Word(_B) * Ratio div 128 + B;

                  if Red   < 0   then Red   := 0;
                  if Red   > $1F then Red   := $1F;
                  if Green < 0   then Green := 0;
                  if Green > $1F then Green := $1F;
                  if Blue  < 0   then Blue  := 0;
                  if Blue  > $1F then Blue  := $1F;

                  RGB2Color(Red,Green,Blue,Data);
                  pDest^ := Data;
               end;
               inc(pSour);
               inc(pDest);
            end;
         end;
      end;
      ceRed   :
      begin
         for y := 0 to Pred(SizeY) do begin
            pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
            pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
            for x := 0 to Pred(SizeX) do begin
               Data := pSour^;
               if Data <> 0 then begin
                  Color2RGB(pDest^,R,G,B);
                  tr1 := (SizeXHalf - Abs(Integer(x) - SizeXHalf));
                  tr2 := (SizeYHalf - Abs(Integer(y) - SizeYHalf));
                  tr3 := tr1 * tr2 div Division;
                  Inc(G,tr3 shr 1);
                  Inc(B,tr3 shr 1);
                  Inc(R,Integer($1F) * Ratio div 64);
                  RGB2Color(R,G,B,Data);
                  pDest^ := Data;
               end;
               inc(pSour);
               inc(pDest);
            end;
         end;
      end;
      ceGreen :
      begin
         for y := 0 to Pred(SizeY) do begin
            pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
            pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
            for x := 0 to Pred(SizeX) do begin
               Data := pSour^;
               if Data <> 0 then begin
                  Color2RGB(pDest^,R,G,B);
                  tr1 := (SizeXHalf - Abs(Integer(x) - SizeXHalf));
                  tr2 := (SizeYHalf - Abs(Integer(y) - SizeYHalf));
                  tr3 := tr1 * tr2 div Division;
                  Inc(B,tr3 shr 1);
                  Inc(R,tr3 shr 1);
                  Inc(G,Integer($1F) * Ratio div 64);
                  RGB2Color(R,G,B,Data);
                  pDest^ := Data;
               end;
               inc(pSour);
               inc(pDest);
            end;
         end;
      end;
      ceBlue  :
      begin
         for y := 0 to Pred(SizeY) do begin
            pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
            pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
            for x := 0 to Pred(SizeX) do begin
               Data := pSour^;
               if Data <> 0 then begin
                  Color2RGB(pDest^,R,G,B);
                  tr1 := (SizeXHalf - Abs(Integer(x) - SizeXHalf));
                  tr2 := (SizeYHalf - Abs(Integer(y) - SizeYHalf));
                  tr3 := tr1 * tr2 div Division;
                  Inc(G,tr3 shr 1);
                  Inc(R,tr3 shr 1);
                  Inc(B,Integer($1F) * Ratio div 64);
                  RGB2Color(R,G,B,Data);
                  pDest^ := Data;
               end;
               inc(pSour);
               inc(pDest);
            end;
         end;
      end;
   end;

   Sour.UnLock(SourDesc.lpSurface);
   Dest.UnLock(DestDesc.lpSurface);
end;

procedure putBrightImage(DestX, DestY : integer;
                         SourRect     : TRect;
                         Sour, Dest   : IDirectDrawSurface7;
                         Ratio        : Integer);
var
   R, G, B                    : Byte;
   Data                       : Word;
   x, y                       : Integer;
   SourX, SourY, SizeX, SizeY : Integer;
   SourDesc, DestDesc         : TDDSurfaceDesc2;
   pSour, pDest               : PWord;
begin

   if (not Assigned(Sour)) or (not Assigned(Dest)) then Exit;

   SourX := SourRect.Left;
   SourY := SourRect.Top;
   SizeX := SourRect.Right  - SourRect.Left;
   SizeY := SourRect.Bottom - SourRect.Top;

   if (DestX < MIN_X_BOUNDARY) or (DestX+SizeX >= MAX_X_BOUNDARY) or
      (DestY < MIN_Y_BOUNDARY) or (DestY+SizeX >= MAX_Y_BOUNDARY) then begin
      PutEffectVBClip(DestX,DestY,SourRect,Sour,Dest,Ratio);
      Exit;
   end;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Sour.Lock(nil,SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Dest.Lock(nil,DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   if Ratio >  64 then Ratio :=  64;
   if Ratio < -64 then Ratio := -64;

   inc(Ratio,64);

   for y := 0 to Pred(SizeY) do begin
      pSour := PWord(Longint(SourDesc.lpSurface)+(SourY+y)*SourDesc.lPitch+(SourX)*BPP);
      pDest := PWord(Longint(DestDesc.lpSurface)+(DestY+y)*DestDesc.lPitch+(DestX)*BPP);
      for x := 0 to Pred(SizeX) do begin
         Data := pSour^;
         if Data <> 0 then begin
            Color2RGB(Data,R,G,B);
            R := Word(R) * Ratio div 64;
            G := Word(G) * Ratio div 64;
            B := Word(B) * Ratio div 64;
            RGB2Color(R,G,B,pDest^);
         end;
         inc(pSour);
         inc(pDest);
      end;
   end;

   Sour.UnLock(SourDesc.lpSurface);
   Dest.UnLock(DestDesc.lpSurface);
end;

procedure TransparentCopy50(DestX, DestY : Integer;
                            SourX, SourY : Integer;
                            Size         : TPoint;
                            Sour, Dest   : IDirectDrawSurface7);
var
   y                   : Integer;
   SourDesc, DestDesc  : TDDSurfaceDesc2;
   pSour, pDest        : PWord;
   MASK_DWORD          : DWORD;
begin

   if (not Assigned(Sour)) or (not Assigned(Dest)) then Exit;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc);
   with Basic do repeat
   until MakeItSo(Sour.Lock(nil,SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc);
   with Basic do repeat
   until MakeItSo(Dest.Lock(nil,DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   if Basic.PixelFormat = pfRGB15 then MASK_DWORD := $3DEF3DEF
                                  else MASK_DWORD := $7BEF7BEF;

   for y := 0 to Pred(Size.Y) do begin
      pSour := PWord(DWORD(SourDesc.lpSurface) + (SourY+y) * SourDesc.lPitch + SourX * BPP);
      pDest := PWord(DWORD(DestDesc.lpSurface) + (DestY+y) * DestDesc.lPitch + DestX * BPP);
      asm
         push  esi
         push  edi

         mov   esi, pSour
         mov   edi, pDest

         mov   ecx, Size.X
         shr   ecx, 1

      @@LOOP:

         lodsd                  // mov eax, [esi]  +  add esi, 4
         shr   eax, 1
         and   eax, MASK_DWORD
         mov   edx, [edi]
         shr   edx, 1
         and   edx, MASK_DWORD
         add   eax, edx
         stosd                  // mov [edi], eax  +  add edi, 4

         loop  @@LOOP

         pop   edi
         pop   esi
      end;
   end;

   Sour.UnLock(SourDesc.lpSurface);
   Dest.UnLock(DestDesc.lpSurface);

end;

procedure ModifyEdge(var x, y : integer; var ARect : TRect; Mode : TSpriteMode);
begin
   if Mode = smCentered then begin
      x := x - Succ(ARect.Right - ARect.Left) div 2;
      y := y - (ARect.Bottom - ARect.Top);
   end;

   if x < 0 then begin
      ARect.Left  := ARect.Left - x;
      x           := 0;
   end;
   if y < 0 then begin
      ARect.Top   := ARect.Top - y;
      y           := 0;
   end;
   if (x + ARect.Right - ARect.Left) > Pred(MAX_X_LINE) then begin
      ARect.Right := MAX_X_LINE + ARect.Left - x;
   end;
   if (y + ARect.Bottom - ARect.Top) > Pred(MAX_Y_LINE) then begin
      ARect.Bottom := MAX_Y_LINE + ARect.Top - y;
   end;
end;

procedure PutSprite(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);
begin
   ModifyEdge(x,y,ARect,Mode);
   with Basic do repeat
   until MakeItSo(BackBuffer.BltFast(x,y,Image[ImageNumber],@ARect,DDBLTFAST_SRCCOLORKEY));
end;

procedure PutSpriteMesh(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode; Ratio : integer);
var
   R, G, B             : Byte;
   i, j                : integer;
   SourDesc, DestDesc  : TDDSurfaceDesc2;
   bSour, bDest        : PByte;
   pSour, pDest        : PWord;
   SourX, SourY        : integer;
   DestX, DestY        : integer;
   SizeX, SizeY        : integer;
begin
   ModifyEdge(x,y,ARect,Mode);

   SourX := ARect.Left;
   SourY := ARect.Top;
   DestX := x;
   DestY := y;
   SizeX := ARect.Right - ARect.Left;
   SizeY := ARect.Bottom - ARect.Top;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Image[ImageNumber].Lock(PRect(nil),SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(BackBuffer.Lock(PRect(nil),DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   if Ratio = 0 then begin
      for j := 0 to Pred(SizeY) do begin
         pSour := Pointer(DWORD(SourDesc.lpSurface) + (SourY+j) * SourDesc.lPitch + SourX * BPP);
         pDest := Pointer(DWORD(DestDesc.lpSurface) + (DestY+j) * DestDesc.lPitch + DestX * BPP);
         for i := 0 to Pred(SizeX) do begin
            if pSour^ <> 0 then
            if (((x+i+(y+j)*2) mod 3) = 0) then begin
               pDest^ := pSour^;
            end;
            Inc(DWORD(pSour),BPP);
            Inc(DWORD(pDest),BPP);
         end;
      end;
   end else begin
      if Ratio >  64 then Ratio :=  64;
      if Ratio < -64 then Ratio := -64;
      inc(Ratio,64);
      for j := 0 to Pred(SizeY) do begin
         pSour := Pointer(DWORD(SourDesc.lpSurface) + (SourY+j) * SourDesc.lPitch + SourX * BPP);
         pDest := Pointer(DWORD(DestDesc.lpSurface) + (DestY+j) * DestDesc.lPitch + DestX * BPP);
         for i := 0 to Pred(SizeX) do begin
            if pSour^ <> 0 then
            if (((x+i+(y+j)*2) mod 3) = 0) then begin
               Color2RGB(pSour^,R,G,B);
               R := Word(R) * Ratio div 64;
               G := Word(G) * Ratio div 64;
               B := Word(B) * Ratio div 64;
               RGB2Color(R,G,B,pDest^);
            end;
            Inc(DWORD(pSour),BPP);
            Inc(DWORD(pDest),BPP);
         end;
      end;
   end;

   Basic.Image[ImageNumber].UnLock(SourDesc.lpSurface);
   Basic.BackBuffer.UnLock(DestDesc.lpSurface);

end;

procedure PutSpriteTrans(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);
var
   i, j                : integer;
   SourDesc, DestDesc  : TDDSurfaceDesc2;
   bSour, bDest        : PByte;
   pSour, pDest        : PWord;
   SourX, SourY        : integer;
   DestX, DestY        : integer;
   SizeX, SizeY        : integer;
   MaskWord            : Word;
begin

   ModifyEdge(x,y,ARect,Mode);

   SourX := ARect.Left;
   SourY := ARect.Top;
   DestX := x;
   DestY := y;
   SizeX := ARect.Right - ARect.Left;
   SizeY := ARect.Bottom - ARect.Top;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Image[ImageNumber].Lock(PRect(nil),SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(BackBuffer.Lock(PRect(nil),DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   if Basic.PixelFormat = pfRGB15 then MaskWord := $3DEF   // 0011 1101
                                  else MaskWord := $7BEF;  // 0111 1011
{
   for y := 0 to Pred(MAX_Y_LINE) do begin
      pDest := PWord(DWORD(DestDesc.lpSurface) + y * DestDesc.lPitch);
      asm
            push esi
            push edi

            xor  ecx, ecx
            mov  cx,  MAX_X_LINE
            shr  ecx, 1
            mov  esi, pDest
            mov  edi, pDest
      @@LOOP:
            lodsd
            shr  eax, 1
            and  eax, MaskDWord1
            stosd
            loop @@LOOP

            pop  edi
            pop  esi
      end;
   end;
}
{
      for j := 0 to Pred(SizeY) do begin
         bSour := Pointer(DWORD(SourDesc.lpSurface) + (SourY+j) * SourDesc.lPitch + SourX * BPP);
         bDest := Pointer(DWORD(DestDesc.lpSurface) + (DestY+j) * DestDesc.lPitch + DestX * BPP);
         for i := 0 to Pred(SizeX) do begin
            if bSour^ <> 0 then
               bDest^ := Min(bSour^+3,((bSour^ div 8) + 1) * 8 - 1);
            Inc(DWORD(bSour),BPP);
            Inc(DWORD(bDest),BPP);
         end;
      end;
}

   for j := 0 to Pred(SizeY) do begin
      pSour := Pointer(DWORD(SourDesc.lpSurface) + (SourY+j) * SourDesc.lPitch + SourX * BPP);
      pDest := Pointer(DWORD(DestDesc.lpSurface) + (DestY+j) * DestDesc.lPitch + DestX * BPP);
      for i := 0 to Pred(SizeX) do begin
         if pSour^ <> 0 then begin
            pDest^ := (pSour^ shr 1) and MaskWord;
         end;
         Inc(DWORD(pSour),BPP);
         Inc(DWORD(pDest),BPP);
      end;
   end;

   Basic.Image[ImageNumber].UnLock(SourDesc.lpSurface);
   Basic.BackBuffer.UnLock(DestDesc.lpSurface);

end;

procedure PutSpriteTrans75(x, y : integer; ImageNumber : integer; ARect : TRect; Mode : TSpriteMode);
var
   i, j                : integer;
   SourDesc, DestDesc  : TDDSurfaceDesc2;
   bSour, bDest        : PByte;
   pSour, pDest        : PWord;
   SourX, SourY        : integer;
   DestX, DestY        : integer;
   SizeX, SizeY        : integer;
   MaskWord1           : Word;
   MaskWord2           : Word;
begin

   ModifyEdge(x,y,ARect,Mode);

   SourX := ARect.Left;
   SourY := ARect.Top;
   DestX := x;
   DestY := y;
   SizeX := ARect.Right - ARect.Left;
   SizeY := ARect.Bottom - ARect.Top;

   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(Image[ImageNumber].Lock(PRect(nil),SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   with Basic do repeat
   until MakeItSo(BackBuffer.Lock(PRect(nil),DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0));

   if Basic.PixelFormat = pfRGB15 then MaskWord1 := $3DEF   // 0011 1101
                                  else MaskWord1 := $7BEF;  // 0111 1011

   if Basic.PixelFormat = pfRGB15 then MaskWord2 := $1CE7   // 0001 1100
                                  else MaskWord2 := $39E7;  // 0011 1001

   for j := 0 to Pred(SizeY) do begin
      pSour := Pointer(DWORD(SourDesc.lpSurface) + (SourY+j) * SourDesc.lPitch + SourX * BPP);
      pDest := Pointer(DWORD(DestDesc.lpSurface) + (DestY+j) * DestDesc.lPitch + DestX * BPP);
      for i := 0 to Pred(SizeX) do begin
         if pSour^ <> 0 then begin
            pDest^ := (pSour^ shr 1) and MaskWord1 + (pSour^ shr 2) and MaskWord2;
         end;
         Inc(DWORD(pSour),BPP);
         Inc(DWORD(pDest),BPP);
      end;
   end;

   Basic.Image[ImageNumber].UnLock(SourDesc.lpSurface);
   Basic.BackBuffer.UnLock(DestDesc.lpSurface);

end;

procedure FadeCopy(Sour, Dest : IDirectDrawSurface7; FadeRatio : integer);
const
   FADE_RATE = 32;
var
   x, y               : Integer;
   R, G, B            : Byte;
   pSour, pDest       : PWord;
   SourDesc, DestDesc : TDDSurfaceDesc2;
begin

   FillChar(SourDesc,sizeof(TDDSurfaceDesc2),0);
   SourDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   Sour.Lock(nil,SourDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0);

   FillChar(DestDesc,sizeof(TDDSurfaceDesc2),0);
   DestDesc.dwSize := SizeOf(TDDSurfaceDesc2);
   Dest.Lock(nil,DestDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT,0);

   if FadeRatio < 0         then FadeRatio := 0;
   if FadeRatio > FADE_RATE then FadeRatio := FADE_RATE;

   for y := 0 to Pred(MAX_Y_LINE) do begin
      pSour := PWord(DWORD(SourDesc.lpSurface) + y * SourDesc.lPitch);
      pDest := PWord(DWORD(DestDesc.lpSurface) + y * DestDesc.lPitch);
      for x := 0 to Pred(MAX_X_LINE) do begin
         Color2RGB(pSour^,R,G,B);
         R := R * FadeRatio div FADE_RATE;
         G := G * FadeRatio div FADE_RATE;
         B := B * FadeRatio div FADE_RATE;
         RGB2Color(R,G,B,pDest^);
         Inc(pSour);
         Inc(pDest);
      end;
   end;

   Sour.UnLock(SourDesc.lpSurface);
   Dest.UnLock(DestDesc.lpSurface);

end;

function  FixColor(Color : Word) : Word;
var
   Temp : Word;
begin
   Temp := Color;
   case Basic.PixelFormat of
      pfRGB15 :
      asm
         mov   ax, Temp
         mov   dx, ax
         and   dx, 0000000000011111b
         shr   ax, 1
         and   ax, 1111111111100000b
         or    ax, dx
         mov   Temp, ax
      end;
      pfBGR16 :
      asm
         mov   ax, Temp
         mov   bx, ax
         mov   dx, ax
         ror   bx, 5
         rol   dx, 5
         and   bx, 1111100000000000b
         and   dx, 0000000000011111b
         and   ax, 0000011111100000b
         or    ax, bx
         or    ax, dx
         mov   Temp, ax
      end;
   end;
   FixColor := Temp;
end;

function  _RGB_15Bit(_R, _G, _B : byte) : Word;
var
   R, G, B : word;
begin
   R := (_R shr 3) shl 10;
   G := (_G shr 3) shl 5;
   B := (_B shr 3);
   _RGB_15Bit := R or G or B;
end;

procedure _Color2RGB_15Bit(Color : Word; var R, G, B : Byte);
var
   _Color : Word;
begin
   _Color := Color;
   R := Byte( _Color shr 10) and $1F;
   G := Byte((_Color shr  5) and $1F);
   B := Byte( _Color and $1F);
end;

procedure _RGB2Color_15Bit(R, G, B : Byte; var Color : Word);
var
   _Color : Word;
begin
   if R > $1F then R := $1F;
   if G > $1F then G := $1F;
   if B > $1F then B := $1F;
   _Color := Word(R);
   _Color := (_Color shl 5) or Word(G);
   _Color := (_Color shl 5) or Word(B);
   Color  := _Color;
end;

procedure _PutPixel_15bit(x, y : longint; color : word);
begin
   if (x < 0) or (y < 0) or (x >= MAX_X_LINE) or (y > MAX_Y_LINE) then exit;

   asm
      mov  ax, color
      mov  dx, ax
      and  dx, 0000000000011111b
      shr  ax, 1
      and  ax, 1111111111100000b
      or   ax, dx
      mov  color, ax
   end;
   if not Basic.FIsLocked then begin
      Basic.LockBackGround;
   end;
   PWord(Longint(Basic.LockDesc.lpSurface)+y*Basic.LockDesc.lPitch+x*BPP)^ := Color;
end;

function  _RGB_16Bit(_R, _G, _B : Byte) : Word;
var
   R, G, B : Word;
begin
   R := (_R shr 3) shl 11;
   G := (_G shr 2) shl 5;
   B := (_B shr 3);
   _RGB_16Bit := R or G or B;
end;

procedure _Color2RGB_16Bit(Color : Word; var R, G, B : Byte);
var
   _Color : Word;
begin
   _Color := Color;
   R := Byte( _Color shr 11) and $1F;
   G := Byte((_Color shr  6) and $1F);
   B := Byte( _Color and $1F);
end;

procedure _RGB2Color_16Bit(R, G, B : Byte; var Color : Word);
var
   _Color : Word;
begin
   if R > $1F then R := $1F;
   if G > $1F then G := $1F;
   if B > $1F then B := $1F;
   _Color := Word(R);
   _Color := (_Color shl 5) or Word(G);
   _Color := (_Color shl 6) or Word(B);
   Color  := _Color;
end;

procedure _PutPixel_16bit(x, y : Longint; Color : Word);
begin
   if (x < 0) or (y < 0) or (x >= MAX_X_LINE) or (y > MAX_Y_LINE) then Exit;

   if not Basic.FIsLocked then begin
      Basic.LockBackGround;
   end;
   PWord(Longint(Basic.LockDesc.lpSurface)+y*Basic.LockDesc.lPitch+x*BPP)^ := Color;
end;

procedure SetWriteMode(PixelFormat : TPixelFormat);
begin
   case PixelFormat of
      pfRGB16, pfBGR16 :
      begin
         MakeRGB     := _RGB_16Bit;
         Color2RGB   := _Color2RGB_16Bit;
         RGB2Color   := _RGB2Color_16Bit;
         PutPixel    := _PutPixel_16bit;
      end;
      pfRGB15 :
      begin
         MakeRGB     := _RGB_15Bit;
         Color2RGB   := _Color2RGB_15Bit;
         RGB2Color   := _RGB2Color_15Bit;
         PutPixel    := _PutPixel_15bit;
      end;
   end;
   MIN_X_BOUNDARY := 0;
   MAX_X_BOUNDARY := MAX_X_LINE;
   MIN_Y_BOUNDARY := 0;
   MAX_Y_BOUNDARY := MAX_Y_LINE
end;

procedure FlipPage;
var
   s        : string;
   Boundary : TRect;
   TimeGap  : Longint;
begin
   if Basic.FFlippingCount = -50 then begin
//      Boundary := Rect(0,0,MAX_X_LINE-19,MAX_Y_LINE-24);
//      ClipCursor(@Boundary);
   end;

   if Basic.FFlippingCount = 0 then begin
      Basic.FStartTickCount := GetTickCount;
   end;

   Inc(Basic.FFlippingCount);

   if Option. CanDisplayPerformance then
   if Basic.FFlippingCount > 0 then begin
      with Basic.BackCanvas, GameSystem do try
         if GetTickCount <> Basic.FStartTickCount then begin
            TimeGap := (GetTickCount-Basic.FStartTickCount);
            if TimeGap > 0 then FlipPerSecond := (Basic.FFlippingCount*1000) / TimeGap
                           else FlipPerSecond := 100;
            if Option.CanModifyWalkFrame then begin
               WALK_FRAME         := Round(FlipPerSecond / 5);
               MAX_FACE_COUNT_SUB := Round(FlipPerSecond / (5 * PERSON_FRAME));
            end;
            Font.Color := RGB(255,255,255);
            Str(FlipPerSecond : 5 : 2,s);
            Textout(0,0,'FPS : '+s);
            Str(WALK_FRAME,s);
            Textout(0,20,'Walk Frame : '+s);
            Str(MAX_FACE_COUNT_SUB,s);
            Textout(0,40,'Face Count : '+s);
         end;
      finally
         Release;
      end;
   end;

   if Option.IsExclusive then begin
      with Basic do repeat
      until MakeItSo(PrimarySurface.Flip(nil,DDFLIP_WAIT));
   end else begin
      Basic.DirectDraw.WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
      with Basic do repeat
         Boundary := Bounds(Left+WIN_MOUSE_X_GAP,Top+WIN_MOUSE_Y_GAP,MAX_X_LINE,MAX_Y_LINE);
      until MakeItSo(PrimarySurface.Blt(@Boundary,BackBuffer,PRect(nil),DDFLIP_WAIT,PDDBltFX(nil)));
   end;

end;

Initialization

   SetWriteMode(pfRGB15);

end.